
import h5py
import numpy as np
import pandas as pd
import pytest

from motrainer.jackknife import JackknifeGPI


@pytest.fixture
def train_data():
    start_year = 2017
    end_year = 2020
    obs_per_year = 20
    x = np.tile(np.linspace(-0.5, 0.5, obs_per_year), end_year - start_year)
    y = x
    df = pd.DataFrame(
        data={"x": x, "y": y},
        index=pd.date_range(
            start=f"1/1/{start_year}",
            end=f"1/1/{end_year}",
            periods=obs_per_year * (end_year - start_year),
        ),
    )
    return df


class TestJacknife:
    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    def test_initialize_gpi(self, train_data):
        gpi = JackknifeGPI(
            train_data, val_split_year=2019, input_list=["x"], output_list=["y"]
        )
        assert gpi.gpi_input["x"] is not None
        assert gpi.gpi_output["y"] is not None

    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    @pytest.fixture
    def test_gpi_results_exist(self, train_data, tmp_path):
        gpi = JackknifeGPI(
            train_data,
            val_split_year=2019,
            input_list=["x"],
            output_list=["y"],
            export_all_years=False,
            outpath=tmp_path,
        )
        gpi.train(
            searching_space={
                "learning_rate": [0.01, 0.02],
                "num_dense_layers": [1, 2],
                "num_input_nodes": [5, 6],
                "num_dense_nodes": [5, 6],
            },
            optimize_space={
                "best_loss": 10,
                "epochs": 1,
                "x0": [0.01, 1, 5, 5, "relu", 64],
            },
            normalize_method="standard",
            training_method="dnn",
            performance_method="rmse",
            verbose=0,
        )
        gpi.export_best()
        assert gpi.apr_perf is not None
        assert gpi.post_perf is not None
        assert gpi.best_train is not None
        assert gpi.best_year in [2017, 2018]

        # test meta_data saved in hdf5 file
        with h5py.File(tmp_path / "best_optimized_model_2017.h5", 'r') as f:
            best_year = f.attrs['best_year']
            input_list = f.attrs['input_list']

        assert best_year == 2017
        assert input_list == gpi.input_list



import h5py
import numpy as np
import pandas as pd
import pytest
from skopt.space import Real

from motrainer.dnn import NNTrain


@pytest.fixture
def train_data():
    start_year = 2017
    end_year = 2018
    obs_per_year = 20
    x = np.tile(np.linspace(-0.5, 0.5, obs_per_year), end_year - start_year)
    y = x
    df = pd.DataFrame(
        data={"x": x, "y": y},
        index=pd.date_range(
            start=f"1/1/{start_year}",
            end=f"1/1/{end_year}",
            periods=obs_per_year * (end_year - start_year),
        ),
    )
    return df


# Expected searching dimensions
EXPECTED_DIMS = {
    "learning_rate",
    "num_dense_layers",
    "num_input_nodes",
    "num_dense_nodes",
    "activation",
    "batch_size",
}


class TestDNN:
    def test_init_train(self, train_data):
        test_train = NNTrain(train_data["x"], train_data["y"])
        assert test_train.train_input.equals(train_data["x"])
        assert test_train.train_output.equals(train_data["y"])
        assert EXPECTED_DIMS.issubset(test_train.dimensions.keys())


class TestUpdateSpace:
    def test_update_space_list(self, train_data):
        test_train = NNTrain(train_data["x"], train_data["y"])
        searching_space = {
            "learning_rate": [0.01, 0.02],
            "num_dense_layers": [1, 2],
        }
        test_train.update_space(**searching_space)

        for key, value in searching_space.items():
            assert key in test_train.dimensions.keys()
            assert value == searching_space[key]

    def test_update_space_skopt(self, train_data):
        nlayer = Real(low=1, high=2, prior="log-uniform", name="num_dense_layers")
        searching_space = {
            "num_dense_layers": nlayer,
        }

        test_train = NNTrain(train_data["x"], train_data["y"])
        test_train.update_space(**searching_space)

        assert "num_dense_layers" in test_train.dimensions.keys()
        assert test_train.dimensions["num_dense_layers"].low == 1
        assert test_train.dimensions["num_dense_layers"].high == 2


class TestOptimize:
    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    @pytest.fixture
    def test_dnn_results_exist(self, tmp_path):
        datain = pd.DataFrame(
            data={"x1": np.linspace(0, 1, 10), "x2": np.linspace(0, 1, 10)}
        )
        dataout = pd.DataFrame(data={"y": np.linspace(0, 1, 100)})
        test_train = NNTrain(datain, dataout)
        test_train.update_space(
            learning_rate=[0.01, 0.02],
            num_dense_layers=[1, 2],
            num_input_nodes=[5, 6],
            num_dense_nodes=[5, 6],
        )
        test_train.optimize(best_loss=10, epochs=1, x0=[0.01, 2, 5, 5, "relu", 64])
        expected_hyperparameters = sorted(
            zip(
                test_train.gp_result.func_vals,
                test_train.gp_result.x_iters,
                strict=True
                )
        )
        test_train.export(f"{tmp_path}/model")
        assert test_train.model is not None

        # load hyperparameters from "model.h5" and assert they are the same
        with h5py.File(f"{tmp_path}/model.h5", 'r') as f:
            param = f.attrs['hyperparameters']
        # convert hyperparameters to a list of tuples
        hyperparameters = eval(param)
        assert hyperparameters[0][0] == expected_hyperparameters[0][0]

    @pytest.mark.filterwarnings("ignore::DeprecationWarning", "ignore::UserWarning")
    def test_dnn_results_exist_lossweight(self):
        datain = pd.DataFrame(
            data={"x1": np.linspace(0, 1, 10), "x2": np.linspace(0, 1, 10)}
        )
        dataout = datain
        test_train = NNTrain(datain, dataout)
        test_train.update_space(
            learning_rate=[0.01, 0.02],
            num_dense_layers=[1, 2],
            num_input_nodes=[5, 6],
            num_dense_nodes=[5, 6],
        )
        test_train.optimize(
            best_loss=10,
            epochs=1,
            training_method="dnn_lossweights",
            loss_weights=[1, 0.5],
            x0=[0.01, 2, 5, 5, "relu", 64],
        )

        assert test_train.model is not None



import logging
from pathlib import Path

import numpy as np
from sklearn.model_selection import LeaveOneOut

from motrainer.dnn import NNTrain
from motrainer.util import normalize, performance

logger = logging.getLogger(__name__)


class JackknifeGPI:
    """GPI object for neuron netowork training using Jackknife resampling method.

    Methods
    -------
    train(searching_space, optimize_space, normalize_method='standard',
          performance_method='rmse', training_method='dnn', verbose=0)
        train neuron network with given method
    export_best
        export the best results in Jackknife process.
    """

    def __init__(self,
                 gpi_data,
                 val_split_year,
                 input_list,
                 output_list,
                 export_all_years=True,
                 outpath='./jackknife_results'):
        """Initialize JackknifeGPI object.

        Parameters
        ----------
        gpi_data : pandas.DataFrame
            DataFrame of a single GPI.
            Each row represents all properties at a certain timestamp.
            Each column represents a time-series of a property.
        val_split_year : int
            Split year of validation. All data after (include) this year will
            be reserved for benchmarking.
        input_list : list of str
            Column names in gpi_data will will be used as input.
        output_list : list of str
            Column names in gpi_data will will be used as output.
        export_all_years : bool, optional
            Switch to export the results of all years, by default True
        outpath : str, optional
            Results exporting path, by default './jackknife_results'
        """
        logger.info('Initializing Jackkinfe trainning:\n'
                    f'val_split_year: {val_split_year}\n'
                    f'input_list: {input_list}\n'
                    f'output_list: {output_list}\n')

        assert not (
            gpi_data.isnull().values.any()), 'Nan value(s) in gpi_data!'

        self.gpi_data = gpi_data
        self.input_list = input_list
        self.output_list = output_list
        self.gpi_input = gpi_data[input_list].copy()
        self.gpi_output = gpi_data[output_list].copy()
        self.val_split_year = val_split_year
        self.export_all_years = export_all_years
        self.outpath = outpath
        Path(self.outpath).parent.mkdir(parents=True, exist_ok=True)

    def train(self,
              searching_space,
              optimize_space,
              normalize_method='standard',
              performance_method='rmse',
              training_method='dnn',
              verbose=0):
        """Train neuron network with Jackknife resampling method.

        Procedures:
        1. Reserve in/output after self.val_split_year for later benchmarking.
        2. From the rest in/output data, leave out one year as validation data.
        3. Perform neuron network training.
        4. Repeat Step 2 and 3 until all years exept benchmarking years have
            been used for validation.
        5. Select the best trainning by best performance.
        6. Perform benchmarking on reserved data.

        Parameters
        ----------
        searching_space : dict
            Arguments of searching space.
        optimize_space : dict
            Arguments of optimazation space.
        normalize_method : str, optional
            Method of normalization. Choose from 'standard' and 'min_max'.
            By default 'standard'
        performance_method : str, optional
            Method of computing performance. Choose from 'rmse', 'mae',
            'pearson' and 'spearman'.
            By default 'rmse'.
        training_method : str, optional
            Traning method selection. Select from 'dnn' or 'dnn_lossweights'.
            By default 'dnn'
        verbose : int, optional
            Control the verbosity.
            By default 0, which means no screen feedback.
        """
        # Data normalization
        logger.debug(f'Normalizing input/output data. Method: {normalize_method}.')
        self.gpi_input[:], scaler_input = normalize(self.gpi_input,
                                                    normalize_method)
        self.gpi_output[:], scaler_output = normalize(self.gpi_output,
                                                      normalize_method)

        # Data split
        input_years = self.gpi_input.index.year
        output_years = self.gpi_output.index.year
        logger.debug(
            f'Spliting Trainning and validation data. \
              Split year: {self.val_split_year}.')
        jackknife_input = self.gpi_input[input_years < self.val_split_year]
        jackknife_output = self.gpi_output[output_years < self.val_split_year]
        vali_input = self.gpi_input[input_years >= self.val_split_year]
        vali_output = self.gpi_output[output_years >= self.val_split_year]
        year_list = jackknife_input.index.year.unique()

        # Jackknife in time
        loo = LeaveOneOut()
        best_perf_sum = None
        for train_index, test_index in loo.split(year_list): # noqa: B007
            this_year = year_list[test_index[0]]

            input_years = jackknife_input.index.year
            output_years = jackknife_output.index.year

            logger.info(f'Jackknife on year: {this_year}.')
            train_input = jackknife_input[input_years != this_year]
            train_output = jackknife_output[output_years != this_year]

            # check if train_input and train_output are empty, raise value error
            if train_input.empty or train_output.empty:
                raise ValueError(
                    'Trainning data is empty. Please check the val_split_year.'
                    )

            test_input = jackknife_input[input_years == this_year]
            test_output = jackknife_output[output_years == this_year]

            # check if test_input and test_output are empty, raise value error
            if test_input.empty or test_output.empty:
                raise ValueError(
                    'Testing data is empty. Please check the val_split_year.'
                    )

            # Execute training
            training = NNTrain(train_input, train_output)

            # Set searching space
            training.update_space(**searching_space)

            # Optimization
            training.optimize(**optimize_space,
                              training_method=training_method,
                              verbose=verbose)

            # TODO: Add warning if no model selected for the year
            if training.model is None:
                logger.warning(f'No best model was found for year: {str(this_year)}.')
                continue

            if self.export_all_years:
                path_model = f'{self.outpath}/all_years/optimized_model_{this_year}'
                training.export(path_model=path_model)

            # find minimum rmse
            # TODO: mae, pearson, spearman
            apr_perf = performance(test_input, test_output, training.model,
                                   performance_method, scaler_output)
            perf_sum = np.nansum(apr_perf)
            if best_perf_sum is None:
                best_perf_sum = perf_sum
            if perf_sum <= best_perf_sum:
                self.apr_perf = apr_perf
                self.post_perf = performance(vali_input, vali_output,
                                             training.model,
                                             performance_method, scaler_output)
                self.best_train = training
                self.best_year = this_year
        logger.info(f'Found best year: {str(self.best_year)}'
                    f'A-priori performance: {self.apr_perf}'
                    f'Post-priori performance: {self.post_perf}')

    def export_best(self, model_name='best_optimized_model'):
        """Export the best results in Jackknife process."""
        logger.info(
            'Exporting model and hyperparameters of '
            f'year {self.best_year} to {self.outpath}'
            )

        if model_name is not None:
            path_model = f'{self.outpath}/{model_name}_{self.best_year}'
        else:
            path_model = f'{self.outpath}/best_optimized_model_{self.best_year}'

        # write metadata
        metedata = {}
        metedata['input_list'] = self.input_list
        metedata['output_list'] = self.input_list
        metedata['best_year'] = int(self.best_year)

        for key in ['lat', 'lon', 'latitude', 'longitude']:
            if key in self.gpi_data.keys():
                metedata[key] = float(self.gpi_data[key].iloc[0])

        self.best_train.export(path_model=path_model, meta_data=metedata)



import numpy as np
import pytest
import xarray as xr

from motrainer import dataset_split, is_splitable, train_test_split
from motrainer.splitter import _regulate_identifier, _validate_train_test_split


@pytest.fixture
def ds_valid():
    nspace = 10
    ntime = 5
    return xr.Dataset(
        data_vars=dict(
            data=(
                ["space", "time"],
                np.arange(nspace * ntime).reshape((nspace, ntime)),
            ),
        ),
        coords=dict(
            space=(["space"], np.arange(nspace)),
            time=(["time"], np.arange(ntime)),
        ),
    )


class TestValidate:
    def test_valid_ds(self, ds_valid):
        assert is_splitable(ds_valid)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_miss_one_dim(self, ds_valid):
        ds_invalid = ds_valid.drop_dims("space")
        assert not is_splitable(ds_invalid)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_redundant_dim(self, ds_valid):
        ds_invalid = ds_valid.expand_dims({"redundant": 3})
        assert not is_splitable(ds_invalid)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_duplicate_coords(self, ds_valid):
        ds_invalid = ds_valid
        ds_invalid["space"] = np.ones(ds_valid.dims["space"])
        assert not is_splitable(ds_invalid)


class TestModelSplit:
    def test_split_space(self, ds_valid):
        identifier = {"space": np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])}  # 4 and 6
        db = dataset_split(ds_valid, identifier)
        list_ds = db.compute()
        space_lens = [len(list_ds[i].space) for i in range(2)]
        assert len(list_ds) == 2
        assert set(space_lens).issubset(
            [4, 6]
        )  # Check the size of separated is 4 and 6, regardless of order

    def test_split_time(self, ds_valid):
        identifier = {"time": np.array([1, 0, 0, 0, 0])}  # 1 and 4
        db = dataset_split(ds_valid, identifier)
        list_ds = db.compute()
        time_lens = [len(list_ds[i].time) for i in range(2)]
        assert len(list_ds) == 2
        assert set(time_lens).issubset(
            [1, 4]
        )  # Check the size of separated is 1 and 4, regardless of order

    def test_split_2d(self, ds_valid):
        identifier = {
            "space": np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0]),  # 4 and 6
            "time": np.array([1, 0, 0, 0, 0]),  # 1 and 4
        }
        db = dataset_split(ds_valid, identifier)
        list_ds = db.compute()
        samples_lens = [len(list_ds[i].samples) for i in range(4)]
        assert len(list_ds) == 4  # In total 4 ds
        assert set(samples_lens).issubset(
            [4, 16, 6, 24]
        )  # Check sample lenth regardless of order
    
    def test_split_2d_one_dim_not_index(self, ds_valid):
        identifier = {
            "space": np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0]),  # 4 and 6
            "time": np.array([1, 0, 0, 0, 0]),  # 1 and 4
        }
        ds = ds_valid.reset_index("space")
        db = dataset_split(ds, identifier)
        list_ds = db.compute()
        samples_lens = [len(list_ds[i].samples) for i in range(4)]
        assert len(list_ds) == 4  # In total 4 ds
        assert set(samples_lens).issubset(
            [4, 16, 6, 24]
        )  # Check sample lenth regardless of order

    def test_split_by_dim(self, ds_valid):
        db_space = dataset_split(ds_valid, "space")
        db_time = dataset_split(ds_valid, "time")
        list_space = db_space.compute()
        list_time = db_time.compute()
        assert len(list_space) == 10
        assert len(list_time) == 5

    def test_split_space_no_coords(self, ds_valid):
        db_no_space_coords = dataset_split(ds_valid.drop_vars("space"), "space")
        db_no_time_coords = dataset_split(ds_valid.drop_vars("time"), "time")
        list_ds_no_space = db_no_space_coords.compute()
        list_ds_no_time = db_no_time_coords.compute()
        assert len(list_ds_no_space) == 10
        assert len(list_ds_no_time) == 5


class TestTrainTestSplit:
    def test_mask(self, ds_valid):
        mask = ds_valid["space"] < 4
        train, test = train_test_split(ds_valid, mask=mask)
        assert train.space.shape[0] == 4
        assert test.space.shape[0] == 6

    def test_mask_reverse(self, ds_valid):
        mask = ds_valid["space"] < 4
        train, test = train_test_split(ds_valid, mask=mask, reverse=True)
        assert train.space.shape[0] == 6
        assert test.space.shape[0] == 4

    def test_split(self, ds_valid):
        train, test = train_test_split(ds_valid, split={"space": 4})
        assert train.space.shape[0] == 4
        assert test.space.shape[0] == 6

    def test_masksplit_reverse(self, ds_valid):
        train, test = train_test_split(ds_valid, split={"space": 4}, reverse=True)
        assert train.space.shape[0] == 6
        assert test.space.shape[0] == 4

    def test_errors(self, ds_valid):
        with pytest.raises(ValueError) as error_msg:
            _ = train_test_split(ds_valid)
            assert "Either mask or split should be specified" in str(error_msg.value)
        with pytest.raises(ValueError):
            _ = train_test_split(
                ds_valid, mask=ds_valid["space"] < 4, split={"space": 4}
            )


class TestValidateIdentifier:
    def test_dict_identifier(self, ds_valid):
        idf = _regulate_identifier(ds_valid, {"space": range(10)})
        assert idf == {"space": range(10)}

    def test_str_identifier(self, ds_valid):
        idf = _regulate_identifier(ds_valid, "space")
        assert idf == {"space": ds_valid["space"].values}

    def test_str_identifier_only_dims(self, ds_valid):
        ds = ds_valid.drop_vars("space")
        idf = _regulate_identifier(ds, "space")
        assert idf == {"space": range(ds.dims["space"])}

    def test_errors(self, ds_valid):
        with pytest.raises(ValueError):
            _ = _regulate_identifier(ds_valid, {})
        with pytest.raises(ValueError):
            _ = _regulate_identifier(
                ds_valid, {"space": range(10), "non_exists": range(10)}
            )
        with pytest.raises(ValueError):
            _ = _regulate_identifier(ds_valid, "non_exists")
        with pytest.raises(NotImplementedError):
            _ = _regulate_identifier(ds_valid, ["space"])


class TestValidateTrainTestSplit:
    def test_split(self):
        _validate_train_test_split({"space": 5})
        _validate_train_test_split({"time": 20})

    def test_errors(self, ds_valid):
        with pytest.raises(ValueError):
            _ = _validate_train_test_split({"space": 5, "time": 5})

        with pytest.raises(ValueError):
            _ = _validate_train_test_split(["space", "time"])



import unittest

import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
from sklearn.linear_model import LinearRegression

from motrainer import util


def init_trainning():
    x = np.linspace(-0.5, 0.5, 100)
    y = np.linspace(-0.5, 0.5, 100)
    datain = np.column_stack((x, x))
    dataout = np.column_stack((y, y))
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Dense(5, input_shape=(2, )))
    model.add(tf.keras.layers.Dense(units=5))
    model.add(tf.keras.layers.Dense(units=2))
    model.compile(optimizer='adam',
                  loss=tf.keras.losses.MeanSquaredError(),
                  metrics=['mae'])
    model.fit(x=datain, y=dataout, epochs=20, verbose=0, validation_split=0.2)
    return model


class TestUtil(unittest.TestCase):
    model = init_trainning()

    def test_performance(self):
        x = np.linspace(-0.5, 0.5, 1000)
        y = np.linspace(-0.5, 0.5, 1000)
        datain = np.column_stack((x, x))
        data_label = np.column_stack((y, y))

        for method in ['rmse', 'mae', 'pearson', 'spearman']:
            perf = util.performance(datain, data_label, TestUtil.model, method)
            self.assertIsNotNone(perf)

    def test_normalize_standard(self):
        x = np.linspace(-0.5, 0.5, 200) + 1
        datain = np.column_stack((x, x))
        data_norm, scaler = util.normalize(datain, 'standard')
        self.assertAlmostEqual(data_norm.mean(), 0.0, 10)
        self.assertIsNotNone(scaler)

    def test_normalize_minmax(self):
        x = np.linspace(-0.5, 0.5, 200) + 1
        datain = np.column_stack((x, x))
        data_norm, scaler = util.normalize(datain, 'min_max')
        self.assertAlmostEqual(data_norm.min(), 0.0, 10)
        self.assertIsNotNone(scaler)


@pytest.fixture
def sklearn_model():
    # create a simple sklearn model
    df = pd.DataFrame({
        'feature': [1, 2, 3, 4, 5],
        'target': [10, 20, 30, 40, 50]
        })
    model = LinearRegression()
    model.fit(df[['feature']], df['target'])
    return model


class TestUtilSklearnIO:

    def test_sklearn_io(self, sklearn_model, tmp_path):
        model_path = tmp_path / 'test.h5'
        util.sklearn_save(sklearn_model, model_path, meta_data={'test': 1})

        # load the model
        model, metadata = util.sklearn_load(model_path)

        # check the model
        assert isinstance(model, LinearRegression)
        # check the metadata
        assert metadata['test'] == 1



import base64
import os
import pickle

import h5py
import numpy as np
import sklearn.preprocessing
from scipy.stats import pearsonr, spearmanr

# Force tensorflow debug logging off, keep only error logging
if "TF_CPP_MIN_LOG_LEVEL" not in os.environ:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


def performance(data_input, data_label, model, method, scaler_output=None):
    """Compute performance of trained neuron netowrk.

    Parameters
    ----------
    data_input : pandas.DataFrame
        Input data.
    data_label : pandas.DataFrame
        Label data.
    model : tf.keras.models
        Trained model to compute performance.
    method : str
        Method to compute
    scaler_output : optional
        Scaler of output, by default None.
        When not None, function will assume that a normalization has been
        performed to output, and will use scaler_output to transform the output
        back to the original scale.

    Returns
    -------
    float or list of float
        Performance value. If the model gives multiple output, the performance
        will be a list.
    """
    predicted = model.predict(data_input)

    # In case multiple outputs, re-arrange to one df
    if isinstance(predicted, list):
        predicted = np.hstack(predicted)

    # Scale back if the data was normalized
    if scaler_output is not None:
        re_predicted = scaler_output.inverse_transform(predicted, 'f')
        re_label = scaler_output.inverse_transform(data_label, 'f')
    else:
        re_predicted = predicted
        re_label = data_label

    difference = re_predicted - re_label
    perf = np.zeros([predicted.shape[1], 1])
    if method == 'rmse':
        for j in range(predicted.shape[1]):
            perf[j, 0] = np.round(np.sqrt(((difference[j])**2).mean()), 5)
    elif method == 'mae':
        for j in range(predicted.shape[1]):
            perf[j, 0] = np.round((difference[j].mean()), 5)
    elif method == 'pearson':
        for j in range(predicted.shape[1]):
            perf[j, 0] = np.round(pearsonr(re_predicted[:, j], re_label[:, j]),
                                  5)[0]
    elif method == 'spearman':
        for j in range(predicted.shape[1]):
            perf[j,
                 0] = np.round(spearmanr(re_predicted[:, j], re_label[:, j]),
                               5)[0]

    return perf


def normalize(data, method):
    """Pre-normalization for input/output.

    Parameters
    ----------
    data : pandas.DataFrAME
        Data to normalize.
    method : str
        Data to normalize. Choose from 'standard' or 'min_max'.

    Returns
    -------
    list
        A list of [data_norm, scaler]. Normalized data and scaler used for
        normalization.

    """
    if method == 'standard':
        scaler = sklearn.preprocessing.StandardScaler()
    elif method == 'min_max':
        scaler = sklearn.preprocessing.MinMaxScaler()
    else:
        raise NotImplementedError

    data_norm = scaler.fit_transform(data)
    return data_norm, scaler


def sklearn_save(model, path_model, meta_data=None):
    """Save sklearn model to hdf5 file.

    Parameters
    ----------
    model : sklearn.model
        Sklearn model to save.
    path_model : str
        Path to save the model.
    meta_data : Dict, optional
        optional. A dict of meta data to save.

    """
    model_bytes = pickle.dumps(model)

    # Encode the bytes as base64
    model_base64 = base64.b64encode(model_bytes)

    with h5py.File(path_model, 'w') as f:
        f.attrs['model'] = model_base64

        if meta_data is not None:
            for key, value in meta_data.items():
                f.attrs[key] = value


def sklearn_load(path_model):
    """Load sklearn model from hdf5 file.

    Parameters
    ----------
    path_model : str
        Path to the model.

    Returns
    -------
    sklearn.model
        Sklearn model.

    """
    with h5py.File(path_model, 'r') as f:
        if 'model' not in f.attrs:
            raise ValueError("No model found in the hdf5 file.")

        model_base64 = f.attrs['model']

        # Decode the bytes
        model_bytes = base64.b64decode(model_base64)

        # Load the model
        model = pickle.loads(model_bytes)

        meta_data = {}
        for key in f.attrs.keys():
            if key != 'model':
                meta_data[key] = f.attrs[key]


    return model, meta_data



import logging

# disable WARNING:absl:Found untraced functions such as _update_step_xla while saving
# see https://github.com/tensorflow/tensorflow/issues/47554
import absl.logging
import h5py
import skopt
import tensorflow as tf
from skopt.space import Categorical, Integer, Real

from motrainer.model import keras_dnn, keras_dnn_lossweight

absl.logging.set_verbosity(absl.logging.ERROR)


logger = logging.getLogger(__name__)


class NNTrain:
    """Neuron Network trainning object.

    Methods
    -------
    update_space(**kwrags)
        Update searching space of tranning
    optimize(best_loss=1, n_calls=15, epochs=100, noise=0.01, n_jobs=-1,
             kappa=5, validation_split=0.2, x0=[1e-3, 1, 4, 13, 'relu', 64],
             training_method='dnn', loss_weights=None, verbose=0):
        Optimize the neuron network within the searching space by given
        optimization settings
    """

    def __init__(self, train_input, train_output):
        """Initialize NNTrain object.

        Parameters
        ----------
        train_input : pandas.DataFrame
            Input data of trainning
        train_output : pandas.DataFrame
            Output data of trainning
        """
        self.train_input = train_input
        self.train_output = train_output
        self.dimensions = {
            'learning_rate':
            Real(low=5e-4,
                 high=1e-2,
                 prior='log-uniform',
                 name='learning_rate'),
            'num_dense_layers':
            Integer(low=1, high=2, name='num_dense_layers'),
            'num_input_nodes':
            Integer(low=2, high=6, name='num_input_nodes'),
            'num_dense_nodes':
            Integer(low=1, high=128, name='num_dense_nodes'),
            'activation':
            Categorical(categories=['relu'], name='activation'),
            'batch_size':
            Integer(low=7, high=365, name='batch_size')
        }
        self.model = None

    def update_space(self, **kwrags):
        """Update searching space of optimization."""
        for key, value in kwrags.items():
            logger.debug(f'Update seaching sapce: {key}={value}')
            # skopt.space instances
            if isinstance(value, Real | Categorical | Integer):
                self.dimensions[key] = value
                self.dimensions[key].name = key

            # Float searching space, e.g. learning rates
            elif any([isinstance(obj, float) for obj in value]):
                assert len(value) == 2
                if any([isinstance(obj, int) for obj in value]):
                    logger.warning(
                        f'Mixed fload/int type found in {key}:{value}. '
                        'The search space will be interpreted as float. '
                        'If this behavior is not desired, try to specify'
                        f'all elements in {key} with the same type.')
                self.dimensions[key] = Real(low=value[0],
                                            high=value[1],
                                            prior='log-uniform',
                                            name=key)

            # Integer searching space, e.g. num of input nodes
            elif all([isinstance(obj, int) for obj in value]):
                assert len(value) == 2
                self.dimensions[key] = Integer(low=value[0],
                                               high=value[1],
                                               name=key)

            # Categorical searching space, e.g. activation
            elif all([isinstance(obj, str) for obj in value]):
                self.dimensions[key] = Categorical(categories=value, name=key)

            else:
                logger.error(
                    f'Do not understand searching space: {key}:{value}.')
                raise NotImplementedError

    def optimize(self,
                 best_loss=1.,
                 n_calls=15,
                 epochs=100,
                 noise=0.01,
                 n_jobs=-1,
                 kappa=5,
                 validation_split=0.2,
                 x0=None,
                 training_method='dnn',
                 loss_weights=None,
                 verbose=0):
        """Optimize the neuron network within the searching space..

        Parameters
        ----------
        best_loss : float, optional
            Threshold of loss, by default 1.
            If the final loss is larger than best_loss, the model won't be
            accepted.
        n_calls : int, optional
            Total number of evaluations, by default 15
        epochs : int, optional
            Number of epochs to train the model, by default 100
        noise : float, optional
            Variance of input, by default 0.01
        n_jobs : int, optional
            Number of cores to run in parallel, by default -1, which sets the
            number of jobs equal to the number of cores.
        kappa : int, optional
            Controls how much of the variance in the predicted values should be
             taken into account. , by default 5
        validation_split : float, optional
            Float between 0 and 1. Fraction of the training data to be used as
            validation data. , by default 0.2
        x0 : list, optional
            Initial input points., by default [1e-3, 1, 4, 13, 'relu', 64]
        training_method : str, optional
            Traning method selection. Select from 'dnn' or 'dnn_lossweights'.
            By default 'dnn'
        loss_weights : list, optional
            Scalar coefficients (Python floats) to weight the loss
            contributions of different model outputs
            By default None, which means equal weights of all outputs
        verbose : int, optional
            Control the verbosity.
            By default 0, which means no screen feedback.
        """
        self.best_loss = best_loss
        self.keras_verbose = verbose
        self.loss_weights = loss_weights

        @skopt.utils.use_named_args(dimensions=list(self.dimensions.values()))
        def func(**dimensions):
            logger.info(f'optimizing with dimensions: {dimensions}')

            # setup model
            earlystop = tf.keras.callbacks.EarlyStopping(
                monitor='val_loss',
                mode='min',
                verbose=self.keras_verbose,
                patience=30)

            if training_method == 'dnn':
                model = keras_dnn(dimensions, self.train_input.shape[1],
                                  self.train_output.shape[1])
                train_output = self.train_output
            elif training_method == 'dnn_lossweights':
                if self.loss_weights is None:
                    self.loss_weights = [1] * self.train_output.shape[1]
                    logger.warning('loss_weights is None.'
                                   f'Using default weights {self.loss_weights}')
                model = keras_dnn_lossweight(dimensions,
                                             self.train_input.shape[1],
                                             self.train_output.shape[1],
                                             self.loss_weights)
                train_output = [
                    self.train_output.iloc[:, i]
                    for i in range(self.train_output.shape[1])
                ]

            # Fit model
            history = model.fit(x=self.train_input,
                                y=train_output,
                                epochs=epochs,
                                batch_size=dimensions['batch_size'],
                                callbacks=[earlystop],
                                verbose=self.keras_verbose,
                                validation_split=validation_split)

            # Get loss
            loss = history.history['val_loss'][-1]
            if loss < self.best_loss:
                self.model = model
                self.best_loss = loss
                self.history = history.history
            del model
            tf.keras.backend.clear_session()
            return loss

        if x0 is None:
            x0 = [1e-3, 1, 4, 13, 'relu', 64]

        self.gp_result = skopt.gp_minimize(func=func,
                                           dimensions=list(
                                               self.dimensions.values()),
                                           n_calls=n_calls,
                                           noise=noise,
                                           n_jobs=n_jobs,
                                           kappa=kappa,
                                           x0=x0)

    def export(self, path_model=None, meta_data=None):
        """Export model, hyperparameters and training metadata."""
        self.model.save(f"{path_model}.h5", save_format='h5')

        hyperparameters = sorted(
            zip(
                self.gp_result.func_vals,
                self.gp_result.x_iters,
                strict=True
                )
        )

        # open the hdf file and write the hyperparameters as attributes
        with h5py.File(f"{path_model}.h5", 'a') as f:
            # list of tuples are not supported by hdf5, so convert to string
            f.attrs['hyperparameters'] = str(hyperparameters)

            # add meta data to the hdf file
            if meta_data is not None:
                for key, value in meta_data.items():
                    f.attrs[key] = value



"""Implementing different types of neural network.

This script is for the implementation of different types of neural network,
including different structures, different loss functions.
"""

import os

import tensorflow as tf

# Force tensorflow debug logging off, keep only error logging
if "TF_CPP_MIN_LOG_LEVEL" not in os.environ:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


def keras_dnn(dimensions, input_shape, output_shape):
    """Deep Neural Network implemented by Keras.

    by default:
    dimension consists of:
    learning_rate, num_dense_layers,num_input_nodes,
                 num_dense_nodes, activation
    dimension['input_shape'] = train_input.shape[1]
    dimension['output_shape'] = train_output.shape[1]
    """
    model = tf.keras.models.Sequential()
    model.add(
        tf.keras.layers.Dense(dimensions['num_input_nodes'],
                              input_shape=(input_shape, ),
                              activation=dimensions['activation']))

    for i in range(dimensions['num_dense_layers']):
        name = f'layer_dense_{i + 1}'
        model.add(
            tf.keras.layers.Dense(dimensions['num_dense_nodes'],
                                  activation=dimensions['activation'],
                                  name=name))
    model.add(tf.keras.layers.Dense(units=output_shape))
    adam = tf.keras.optimizers.Adam(learning_rate=dimensions['learning_rate'])
    model.compile(optimizer=adam,
                  loss=tf.keras.losses.MeanSquaredError(),
                  metrics=['mae', 'acc'])

    return model


def keras_dnn_lossweight(dimensions, input_shape, output_shape, loss_weights):
    """Deep Neural Network implemented by Keras.

    Implemented to adapt 'loss_weights'.
    """
    inputs = tf.keras.Input(shape=(input_shape, ))

    for i in range(dimensions['num_dense_layers']):
        name = f'layer_dense_{i + 1}'
        if i == 0:
            hidden = tf.keras.layers.Dense(dimensions['num_input_nodes'],
                                           activation=dimensions['activation'],
                                           name=name)(inputs)
            hidden_prev = hidden
        else:
            hidden = tf.keras.layers.Dense(dimensions['num_dense_nodes'],
                                           activation=dimensions['activation'],
                                           name=name)(hidden_prev)
            hidden_prev = hidden

    outputs = []
    for i in range(output_shape):
        name = f'out{i + 1}'
        outputs.append(tf.keras.layers.Dense(1, name=name)(hidden))

    adam = tf.keras.optimizers.Adam(learning_rate=dimensions['learning_rate'])

    model = tf.keras.models.Model(inputs=inputs, outputs=outputs)
    model.compile(optimizer=adam,
                  loss=tf.keras.losses.MeanSquaredError(),
                  metrics=['mae', 'acc'],
                  loss_weights=loss_weights)

    return model



import warnings

import dask.bag as db
import numpy as np
import xarray as xr

MOT_DIMS = ["space", "time"]  # Expected xr.Dataset dimensions


def is_splitable(ds: xr.Dataset) -> bool:
    """Check if a Dataset is can be splitted using MOTrainer.

    The following checks will be applied:
        - The Dastaset has exactly 2 dimensions
        - The 2 dims are "space" and "time"
        - There are no duplicated coordinates
    A UserWarning will be raised for each failed check.

    Parameters
    ----------
    ds : xr.Dataset
        Xarray Dataset to be splitted.

    Returns
    -------
    bool
       Result of check in Boolean.
       If all checks are passed, it will be True. Otherwise False.
    """
    flag_valid = True

    # Dimension size should be 2
    if len(ds.dims) != 2:
        warnings.warn(
            'Dataset should have two dimensions: "space" and "time"',
            UserWarning,
            stacklevel=2,
        )
        flag_valid = False

    # space and time dimensions should exist
    for dim in MOT_DIMS:
        if dim not in ds.dims:
            warnings.warn(
                f"{dim} not found in the dimensions", UserWarning, stacklevel=2
            )
            flag_valid = False

    # Check duplicated coordinates
    for coord in ds.coords:
        if np.unique(ds[coord]).shape != ds[coord].shape:
            warnings.warn(
                f"Duplicated coordinates found in {coord}", UserWarning, stacklevel=2
            )
            flag_valid = False

    return flag_valid


def dataset_split(ds: xr.Dataset, identifier: dict | str) -> db:
    """Split a Dataset by indentifier for independent training tasks.

    Parameters
    ----------
    ds : xr.Dataset
        Xarray Dataset to be splitted.
    identifier : dict | str
        When `indentifier` is a dictionary, its keys should be a subset of
        {"space", "time"},  and map "space" and/or "time" dimension with corresponding
        separation indentifier.

        When `indentifier` is a string, the separation will depends on if `indentifier`
        is a key of coords/data_vars or a dimension name ("space" or "time").
        In the former case the corresponding coords/data_vars will be used as separation
        indentifier.
        In the latter case ds will be separated per entry in that dimension.

    Returns
    -------
    dask.bag
        A Dask Databag of splited Datasets
    """
    identifier = _regulate_identifier(ds, identifier)

    list_id = []
    for key in MOT_DIMS:
        if key in identifier.keys():
            key_id = key + "_id"
            ds = ds.assign_coords({key_id: (key, identifier[key])})
            list_id.append(key_id)

    # Get the name of attributes to groupby
    if len(list_id) > 1:
        # Use space time cls coordinates as multi index
        # Must stack on MOT_DIMS to reduce dims of data variable
        multi_idx = ds.stack(samples=list_id).samples.values
        for dim in MOT_DIMS:
            if dim in ds.indexes:
                ds = ds.reset_index(dim)
        ds = (
            ds.stack(
                samples=MOT_DIMS, create_index=False
            )  # No index creation since this will be added next.
            .assign_coords(samples=multi_idx)
            .set_xindex(list_id)
        )
        key_gb = "samples"
    else:
        key_gb = list_id[0]

    # Groupby and separate to Dask Databags
    list_db = []
    for grp in list(ds.groupby(key_gb)):
        list_db.append(grp[1])
    bags = db.from_sequence(list_db)

    return bags


def train_test_split(
    ds: xr.Dataset,
    mask: xr.DataArray = None,
    split: dict = None,
    reverse: bool = False,
) -> tuple[xr.Dataset, xr.Dataset]:
    """Split data to train and test datasets.

    The split is performed either 1) by specifying the training data mask (`mask`)
    where training data locations are True, or 2) by a specifying a coordinate value
    (`split`) splitting the data into two.

    Parameters
    ----------
    ds : xr.Dataset
        Xarray dataset to split
    mask : xr.DataArray, optional
        Mask, True at training data locations. By default None
    split : dict, optional
        coordinate diactionary in {NAME: coordinates} which split the Dataset into two.
        The part smaller than it will be training, by default None.
    reverse : bool, optional
        Reverse the split results, by default False

    Returns
    -------
    tuple[xr.Dataset, xr.Dataset]
        Split results. In (training, test).

    Raises
    ------
    ValueError
        When neither mask nor split is specified.
    ValueError
        When both mask and split are specified.
    """
    if mask is None and split is None:
        raise ValueError("Either mask or split should be specified.")
    elif mask is not None and split is not None:
        raise ValueError("Only one of mask and split should be specified.")

    # Convert split to mask
    if split is not None:
        _validate_train_test_split(split)
        mask = ds[list(split.keys())[0]] < list(split.values())[0]

    train = ds.where(mask, drop=True)
    test = ds.where(~mask, drop=True)

    return (test, train) if reverse else (train, test)


def _regulate_identifier(ds: xr.Dataset, identifier: dict | str) -> dict:
    """Regulate the split identifier w.r.t. the dataset.

    When `indentifier` is a dictionary:
        function will check if the keys are a subset of {"space", "time"}.

    When `indentifier` is a string:
        if `indentifier` is a key of any coords/data_vars, the corresponding
        coords/data_vars will be converter to a dict and used as indentifier
        if `indentifier` is not a key of any coords/data_vars, but is one of dimensions
        i.e. "space" or "time", then it will be converted to either:
            {"space": range(ds.dims["space"])} or
            {"time": range(ds.dims["space"])}

    Parameters
    ----------
    ds : xr.Dataset
        referce DS
    identifier : dict | str
        input indertifier

    Returns
    -------
    dict
        regulated identifier

    Raises
    ------
    ValueError
        identifier is an empty dict.
    ValueError
        identifier is a dict but has keys other that "space" and "time".
    ValueError
        identifier is a str but does not match any key in dims, coords or data_vars.
    NotImplementedError
        identifier is not a dict nor str
    """
    if isinstance(identifier, dict):
        if len(identifier.keys()) == 0:
            raise ValueError("identifier is empty")
        if not set(identifier.keys()).issubset(MOT_DIMS):
            raise ValueError('Acceptable keys are "space" and/or "time".')
    elif isinstance(identifier, str):
        if identifier in ds.variables:
            identifier = {identifier: ds[identifier].values}
        elif identifier in ds.dims:
            identifier = {identifier: range(ds.dims[identifier])}
        else:
            raise ValueError(f'Cannot find "{identifier}" in the Dataset')
    else:
        raise NotImplementedError("identifier must be a dictionary or string")

    return identifier


def _validate_train_test_split(split):
    if not isinstance(split, dict):
        raise ValueError("split should be a dict")

    if len(split.keys()) != 1:
        raise ValueError("split should only have 1 key")

    return None
